/** @namespace req.body.passwordConfirmation */

const crypto = require('crypto');
const util = require('util');
const uuid = require('uuid/v4');
const validator = require('validator');
const pbkdf2 = util.promisify(crypto.pbkdf2);
const { handleRestfulGetRequest, handleRestfulPostRequest, handleRestfulPutRequest, handleRestfulPatchRequest, handleRestfulDeleteRequest } = require('./rest_handlers');

const ITERATIONS = 10000;
const KEYLEN = 64;
const DIGEST = 'sha512';

/*
cases

sign up
sign in
forgot password
update profile
change password
change email
delete account (clean up docs?)
 */

exports.handleUserRequests = async (req, res, next) => {
  if (req.resource === 'user') {
    await handleGetRequest(req, res);
    await handlePostRequest(req, res, next);
    await handlePutRequest(req, res, next);
    await handlePatchRequest(req, res, next);
    await handleDeleteRequest(req, res);
  }
};

const handleGetRequest = async (req, res) => {
  await handleRestfulGetRequest(req, res, transformMany);
};

const handlePostRequest = async (req, res, next) => {
  if (req.method === 'POST') {
    const { email, password, passwordConfirmation } = req.body;
    req.body = sanitizeBody(req);

    const fieldErrors = validateUpsert(req.db, email, password, passwordConfirmation);
    if (fieldErrors.length > 0) {
      return next(fieldErrors);
    }

    const salt = crypto.randomBytes(128).toString('base64');
    const derivedKey = await pbkdf2(password, salt, ITERATIONS, KEYLEN, DIGEST);
    const passwordHash = derivedKey.toString('hex');
    const apiKey = uuid();
    const apiSecretKey = uuid();
    req.body = Object.assign({}, req.body, { salt, iterations: ITERATIONS, keylen: KEYLEN, digest: DIGEST, passwordHash, apiKey, apiSecretKey });

    await handleRestfulPostRequest(req, res, transformOne);
    // TODO: send email
  }
};

const handlePutRequest = async (req, res, next) => {
  if (method === 'PUT') {
    const { email, password, passwordConfirmation } = req.body;
    req.body = sanitizeBody(req);

    const fieldErrors = validateUpsert(req.db, email, password, passwordConfirmation);
    if (fieldErrors.length > 0) {
      return next(fieldErrors);
    }

    await handleRestfulPutRequest(req, res)
    // TODO: send email
  }
};

const handlePatchRequest = async (req, res, next) => {
  if (method === 'PATCH') {
    const { email, password, passwordConfirmation } = req.body;
    req.body = sanitizeBody(req);

    const fieldErrors = validatePatch(req.db, email, password, passwordConfirmation);
    if (fieldErrors.length > 0) {
      return next(fieldErrors);
    }

    await handleRestfulPatchRequest(req, res)
    // TODO: send email (if email or password is changed)
  }
};

const handleDeleteRequest = async (req, res) => {
  await handleRestfulDeleteRequest(req, res);
};

const sanitizeBody = (req) => {
  const body = Object.assign({}, req.body);
  if (!body.email) {
    delete body.email;
  }
  delete body.password;
  delete body.passwordConfirmation;
  delete body.salt;
  delete body.iterations;
  delete body.keylen;
  delete body.digest;
  delete body.passwordHash;
  return body;
};

const validateUpsert = async (db, email, password, passwordConfirmation) => {
  const fieldErrors = [];
  if (!email) {
    fieldErrors.push({ code: -1001, field: 'email', message: 'Email is required' });
  }
  if (email && !validator.isEmail(email)) {
    fieldErrors.push({ code: -1005, field: 'email', message: 'Email is invalid' });
  }
  if (!password) {
    fieldErrors.push({ code: -1002, field: 'password', message: 'Password is required' });
  }
  if (!passwordConfirmation) {
    fieldErrors.push({ code: -1003, field: 'passwordConfirmation', message: 'Password confirmation is required' });
  }
  if (password && passwordConfirmation && password !== passwordConfirmation) {
    fieldErrors.push({ code: -1004, field: 'passwordConfirmation', message: 'Passwords don\'t match' });
  }
  if (email && await db.collection(req.resource).countDocuments({ email }) !== 0) {
    fieldErrors.push({ code: -1000, field: 'email', message: 'Email already exists' })
  }
  return fieldErrors;
};

const validatePatch = async (db, email, password, passwordConfirmation) => {
  const fieldErrors = [];
  if (email && !validator.isEmail(email)) {
    fieldErrors.push({ code: -1005, field: 'email', message: 'Email is invalid' });
  }
  if (password && passwordConfirmation && password !== passwordConfirmation) {
    fieldErrors.push({ code: -1004, field: 'passwordConfirmation', message: 'Passwords don\'t match' });
  }
  if (email && await db.collection(req.resource).countDocuments({ email }) !== 0) {
    fieldErrors.push({ code: -1000, field: 'email', message: 'Email already exists' })
  }
  return fieldErrors;
};

const transformMany = (entities) => {
  return entities.map(e => transformOne(e));
};

const transformOne = (doc) => {
  doc.id = doc._id;
  delete doc._id;
  delete doc.salt;
  delete doc.iterations;
  delete doc.keylen;
  delete doc.digest;
  delete doc.passwordHash;
  return doc;
};
